@@ -433,7 +433,8 @@ create <- function(x, ...) {
 #'
 #' @param x a relational data object, into which to insert data, such as a
 #'   \code{\link{relation}} or \code{\link{database}} object.
-#' @param vals a data frame, containing data to insert.
+#' @param vals a data frame, containing data to insert. Column names must be
+#'   unique.
 #' @param relations a character vector, containing names of elements of \code{x}
 #'   into which to insert data. By default, \code{insert} attempts to insert
 #'   data into every element.
@@ -441,12 +442,30 @@ create <- function(x, ...) {
 #'   all attributes of all elements of \code{x[relations]}. By default, it is
 #'   not, and data is only inserted into elements of \code{x[relations]} whose
 #'   attributes are all present in \code{vals}.
+#' @param keep_rownames a logical or a string, indicating whether to include the
+#'   row names as a column. If a string is given, it is used as the name for the
+#'   column, otherwise the column is named "row". Set to FALSE by default.
+#' @param digits a positive integer, indicating how many significant digits are
+#'   to be used for numeric and complex variables. A value of \code{NA} results
+#'   in no rounding. By default, this uses \code{getOption("digits")}, similarly
+#'   to \code{\link{format}}. See the "Floating-point variables" section for
+#'   \code{\link{discover}} for why this rounding is necessary for consistent
+#'   results across different machines. See the note in
+#'   \code{\link{print.default}} about \code{digits >= 16}.
 #' @param ... further arguments pass on to methods.
 #'
 #' @return An R object of the same class as \code{x}, containing the additional
 #'   new data.
 #' @export
-insert <- function(x, vals, relations = names(x), all = FALSE, ...) {
+insert <- function(
+  x,
+  vals,
+  relations = names(x),
+  all = FALSE,
+  keep_rownames = FALSE,
+  digits = getOption("digits"),
+  ...
+) {
   UseMethod("insert")
 }
 
