@@ -589,22 +589,11 @@ describe("relation_schema", {
     )
   })
   it("concatenates without losing attribute orderings, if consistent", {
-    concatenate_keeps_attribute_order <- function(...) {
-      lst <- list(...)
-      expect_silent(res <- c(...))
-      for (index in seq_along(lst)) {
-        expect_identical(
-          attrs_order(lst[[!!index]]),
-          intersect(attrs_order(res), attrs_order(lst[[!!index]]))
-        )
-      }
-    }
-
     forall(
       gen.sample(letters[1:8], gen.element(1:3)) |>
         gen.with(
           sort %>>%
-            with_args(relation_schema, schema = setNames(list(), character()))
+            with_args(relation_schema, schemas = setNames(list(), character()))
         ) |>
         gen.list(from = 2, to = 5),
       concatenate_keeps_attribute_order,
